'''
This is the torch code 
'''

import torch
from pytorch3d.ops import knn_points

def query(q_pts:torch.tensor, verts:torch.tensor, tets:torch.tensor, indexes:torch.tensor, K=4) -> tuple:
    tet_verts = verts[tets]  # [NT,4,3]
    tet_verts_center = tet_verts.mean(dim=1)  # [NT,3]

    knn_res = knn_points(q_pts.unsqueeze(0), tet_verts_center.unsqueeze(0), K=K)
    neighbor_idx = knn_res.idx[0]  # [Nq, K], the closest tet center to the query points
    neighbor_tets = tet_verts[neighbor_idx]  # [Nq,K,4,3]

    ##### TODO include the tetrehedra that containing the nearest vertex of query point
    
    ### start calculation. 
    det_verts = torch.cat([q_pts[:,None,None,:].expand(-1,K,-1,-1), neighbor_tets], dim=2)  # [Nq,K,5,3]
    # print("det_verts:", det_verts)
    ### Need to calculate 5 determinants for barycentric calculation
    det_verts = det_verts[:,:,None,:,:].expand(-1,-1,5,-1,-1)  # [Nq,K,5,5,3]
    indexes_shape = det_verts.shape[:2] + indexes.shape
    indexes = indexes[None,None,...].expand(*indexes_shape)  # [Nq,K,5,4,3]
    dets = torch.gather(det_verts, index=indexes, dim=3)  # [Nq,K,5,4,3]
    dets = torch.cat([dets, torch.ones_like(dets[...,:1])], dim=-1)  # [Nq,K,5,4,4]

    dets_val = torch.linalg.det(dets)  # [Nq,K,5]

    uvwz = dets_val[...,1:] / (dets_val[...,0:1] + 1e-9) # [Nq,K,4]
    u, v, w, z = uvwz[...,0], uvwz[...,1], uvwz[...,2], uvwz[...,3] # [Nq,K]
    in_tet_mask = torch.logical_and(u > 0, v > 0)
    in_tet_mask = torch.logical_and(in_tet_mask, w > 0)
    in_tet_mask = torch.logical_and(in_tet_mask, z > 0) # [Nq,K]

    in_tet_mask = torch.cat([in_tet_mask.int(), torch.ones_like(in_tet_mask[:,:1])], dim=-1) # [Nq,K+1]
    tet_idx = torch.argmax(in_tet_mask, dim=1) # [Nq,]
    uvwz = torch.cat([uvwz, torch.zeros_like(uvwz[:,:1,:])], dim=1)  # [Nq,K+1,4]
    # if not contained in any tets, set index and barys to -1, [0,0,0,0]
    barycentric = torch.gather(uvwz, dim=1, index=tet_idx[:,None,None].expand(-1,1,4)).permute(0,2,1)  # [Nq,4,1]

    ### tet_idx is the in local index in the K tets. Need to change to global index
    neighbor_idx = torch.cat([neighbor_idx, -1 * torch.ones_like(neighbor_idx[:,:1])], dim=1)  # [Nq,K+1]
    tet_idx = torch.gather(neighbor_idx, index=tet_idx[:,None], dim=1)  # [Nq, 1]
    return tet_idx.squeeze(-1), barycentric.squeeze(-1)


import torch
from query_cpp import query

def genIndex(use_cuda=True) -> torch.tensor:
    indexes = []
    index = torch.tensor([[1,2,3,4]],dtype=torch.long).T.expand(4,3)
    indexes.append(index.unsqueeze(0))
    index = torch.tensor([[0,2,3,4]],dtype=torch.long).T.expand(4,3)
    indexes.append(index.unsqueeze(0))
    index = torch.tensor([[1,0,3,4]],dtype=torch.long).T.expand(4,3)
    indexes.append(index.unsqueeze(0))
    index = torch.tensor([[1,2,0,4]],dtype=torch.long).T.expand(4,3)
    indexes.append(index.unsqueeze(0))
    index = torch.tensor([[1,2,3,0]],dtype=torch.long).T.expand(4,3)
    indexes.append(index.unsqueeze(0))
    indexes = torch.cat(indexes, dim=0) # [5,4,3]
    if use_cuda:
        indexes = indexes.cuda()
    return indexes


def cal_barycentric(q_pts:torch.tensor, verts:torch.tensor, tets:torch.tensor, K=4, use_cuda=True):
    """_summary_

    Args:
        q_pts (torch.tensor): the query points, [Nq,3]
        verts (torch.tensor): the vertices tethedra, [NV,3]
        tets (torch.tensor): the vertex index of each tet, [NT,4]
        K (int): the nearest neighbors for searching to accelerate query
        use_cuda (bool, optional): _description_. Defaults to True.

    Returns:
        tuple: the tet_idx and corresponding barycentric coordinates
    """
    if use_cuda:
        q_pts, verts, tets = q_pts.cuda(), verts.cuda(), tets.cuda()
    indexes = genIndex(use_cuda=use_cuda)
    res = query(q_pts, verts, tets, indexes, 4)
    return res



"""author: SunYangtian
   time: 2022/06/07
"""

from gzip import READ
from setuptools import find_packages, setup, Extension
from torch.utils import cpp_extension

with open("./README.md", 'r', encoding='utf-8') as f:
      README = f.read()

setup(name='query_pts',
      version='0.2.1',
      author="SunYangtian",
      author_email="sunyangtian98@gmail.com",
      description="A package for calculate barycentric coordinates in the tetrahedras.",
      long_description=README,
      long_description_content_type="text/markdown",
      packages=find_packages(where='src'),
      package_dir={"": "src"},
      install_requires=[
            'torch>=1.7.1',],
      ext_modules=[cpp_extension.CUDAExtension('query_cpp', 
                    ["src/query_pts/cpp_extension/cpp_query.cpp", "src/query_pts/cpp_extension/knn.cu"],)],
      cmdclass={'build_ext': cpp_extension.BuildExtension})

### query_pts.cuda: the name of .egg, and the name for pip install
### query_cpp: the name of cpp build file

# >>>> equivalent code >>>>
# Extension(
#    name='lltm_cpp',
#    sources=['lltm.cpp'],
#    include_dirs=cpp_extension.include_paths(),
#    language='c++')